@@ -25,8 +25,8 @@ from collections import defaultdict
 
 import attr
 
-from hypothesis import settings as Settings
 from hypothesis import Phase, Verbosity, HealthCheck
+from hypothesis import settings as Settings
 from hypothesis.reporting import debug_report
 from hypothesis.internal.compat import Counter, ceil, hbytes, hrange, \
     int_to_text, int_to_bytes, benchmark_time, int_from_bytes, \
@@ -35,7 +35,7 @@ from hypothesis.utils.conventions import UniqueIdentifier
 from hypothesis.internal.healthcheck import fail_health_check
 from hypothesis.internal.conjecture.data import MAX_DEPTH, Status, \
     StopTest, ConjectureData
-from hypothesis.internal.conjecture.minimizer import minimize
+from hypothesis.internal.conjecture.minimizer import minimize, minimize_int
 
 # Tell pytest to omit the body of this module from tracebacks
 # http://doc.pytest.org/en/latest/example/simple.html#writing-well-integrated-assertion-helpers
@@ -1220,12 +1220,16 @@ class Shrinker(object):
 
         # We keep track of the current best example on the shrink_target
         # attribute.
+        self.shrink_target = None
         self.update_shrink_target(initial)
 
     def incorporate_new_buffer(self, buffer):
         buffer = hbytes(buffer[:self.shrink_target.index])
         assert sort_key(buffer) < sort_key(self.shrink_target.buffer)
 
+        if self.shrink_target.buffer.startswith(buffer):
+            return False
+
         if not self.__engine.prescreen_buffer(buffer):
             return False
 
@@ -1287,31 +1291,25 @@ class Shrinker(object):
 
         """
 
-        # This will automatically be run during the normal loop, but it's worth
-        # running once before the coarse passes so they don't spend time on
-        # useless data.
-        self.remove_discarded()
-
-        # We only run this once because delta-debugging is essentially useless
-        # and often actively harmful when run on examples which are "close to
-        # shrunk" - the attempts to delete long runs mostly just don't work,
-        # and we spend a huge amount of time trying them anyway. The
-        # greedy_interval_deletion pass will achieve the same end result but
-        # is much cheaper when run on close to shrunk examples.
-        self.delta_interval_deletion()
-
         run_expensive_shrinks = False
 
         prev = None
         while prev is not self.shrink_target:
             prev = self.shrink_target
+
+            # We reset our tracking of what changed at the beginning of the
+            # loop so that we don't get distracted by things that change once
+            # and then are stable thereafter.
+            self.clear_change_tracking()
+
             self.remove_discarded()
+            self.adaptive_example_deletion()
             self.zero_draws()
             self.minimize_duplicated_blocks()
             self.minimize_individual_blocks()
             self.reorder_blocks()
-            self.greedy_interval_deletion()
             self.lower_dependent_block_pairs()
+            self.lower_common_block_offset()
 
             # Passes after this point are expensive: Prior to here they should
             # all involve no more than about n log(n) shrinks, but after here
@@ -1482,6 +1480,67 @@ class Shrinker(object):
             t.buffer[:t.blocks[i][0]] in self.__shrinking_prefixes
         )
 
+    def lower_common_block_offset(self):
+        """Sometimes we find ourselves in a situation where changes to one part
+        of the byte stream unlock changes to other parts. Sometimes this is
+        good, but sometimes this can cause us to exhibit exponential slow
+        downs!
+
+        e.g. suppose we had the following:
+
+        m = draw(integers(min_value=0))
+        n = draw(integers(min_value=0))
+        assert abs(m - n) > 1
+
+        If this fails then we'll end up with a loop where on each iteration we
+        reduce each of m and n by 2 - m can't go lower because of n, then n
+        can't go lower because of m.
+
+        This will take us O(m) iterations to complete, which is exponential in
+        the data size, as we gradually zig zag our way towards zero.
+
+        This can only happen if we're failing to reduce the size of the byte
+        stream: The number of iterations that reduce the length of the byte
+        stream is bounded by that length.
+
+        So what we do is this: We keep track of which blocks are changing, and
+        then if there's some non-zero common offset to them we try and minimize
+        them all at once by lowering that offset.
+
+        This may not work, and it definitely won't get us out of all possible
+        exponential slow downs (an example of where it doesn't is where the
+        shape of the blocks changes as a result of this bouncing behaviour),
+        but it fails fast when it doesn't work and gets us out of a really
+        nastily slow case when it does.
+
+        """
+        if len(self.__changed_blocks) <= 1:
+            return
+
+        self.debug('Removing common block offset')
+
+        current = self.shrink_target
+
+        blocked = [current.buffer[u:v] for u, v in current.blocks]
+
+        changed = sorted(self.__changed_blocks)
+
+        ints = [int_from_bytes(blocked[i]) for i in changed]
+        offset = min(ints)
+        if offset == 0:
+            return
+
+        for i in hrange(len(ints)):
+            ints[i] -= offset
+
+        def reoffset(o):
+            new_blocks = list(blocked)
+            for i, v in zip(changed, ints):
+                new_blocks[i] = int_to_bytes(v + o, len(blocked[i]))
+            return self.incorporate_new_buffer(hbytes().join(new_blocks))
+
+        minimize_int(offset, reoffset)
+
     def mark_shrinking(self, blocks):
         """Mark each of these blocks as a shrinking block: That is, lowering
         its value lexicographically may cause less data to be drawn after."""
@@ -1493,8 +1552,27 @@ class Shrinker(object):
             prefix = t.buffer[:t.blocks[i][0]]
             self.__shrinking_prefixes.add(prefix)
 
+    def clear_change_tracking(self):
+        self.__changed_blocks.clear()
+
     def update_shrink_target(self, new_target):
         assert new_target.frozen
+        if self.shrink_target is not None:
+            if new_target.blocks != self.shrink_target.blocks:
+                self.__changed_blocks = set()
+            else:
+                current = self.shrink_target.buffer
+                new = new_target.buffer
+
+                for i, (u, v) in enumerate(self.shrink_target.blocks):
+                    if (
+                        i not in self.__changed_blocks and
+                        current[u:v] != new[u:v]
+                    ):
+                        self.__changed_blocks.add(i)
+        else:
+            self.__changed_blocks = set()
+
         self.shrink_target = new_target
         self.__shrinking_block_cache = {}
         self.__intervals = None
@@ -1632,7 +1710,7 @@ class Shrinker(object):
         doing rejection sampling - e.g. as a result of an integer range, or a
         filtered strategy.
 
-        Such data will also be handled by the greedy_interval_deletion pass,
+        Such data will also be handled by the adaptive_example_deletion pass,
         but that pass is necessarily more conservative and will try deleting
         each interval individually. The common case is that all data drawn and
         rejected can just be thrown away immediately in one block, so this pass
@@ -1663,54 +1741,22 @@ class Shrinker(object):
         # will reset the flag if the status changes.
         self.__discarding_failed = not self.incorporate_new_buffer(attempt)
 
-    def delta_interval_deletion(self):
-        """Attempt to delete every interval in the example, but be more
-        aggressive about it than we are in greedy_interval_deletion.
+    def adaptive_example_deletion(self):
+        """Attempt to delete every draw call, plus some short sequences of draw
+        calls.
 
-        The idea here is that if we have a long list of, say, tens or
-        hundreds of elements, we *could* try deleting each element of the list
-        individually, but if only a handful of elements of the list matter then
-        this is a bit of a waste of time. Wouldn't it be better if we deleted
-        many of them at once?
-
-        The approach we use is approximately delta-debugging from Hildebrandt
-        and Zeller's "Simplifying failure-inducing input" - we try deleting
-        progressively shorter runs, eventually bottoming out in deleting single
-        intervals as per greedy_interval_deletion.
-
-        Note that this is *not* the same as actual delta-debugging on the the
-        buffer, because we are deleting intervals instead of bytes. These may
-        overlap with each-other.
-
-        """
-
-        self.debug('delta interval deletes')
-
-        k = len(self.intervals) // 2
-        while k > 0:
-            i = 0
-            while i + k <= len(self.intervals):
-                bitmask = [True] * len(self.shrink_target.buffer)
-
-                for u, v in self.intervals[i:i + k]:
-                    for t in range(u, v):
-                        bitmask[t] = False
-
-                if not self.incorporate_new_buffer(hbytes(
-                    b for b, v in zip(self.shrink_target.buffer, bitmask)
-                    if v
-                )):
-                    i += k
-            k //= 2
-
-    def greedy_interval_deletion(self):
-        """Attempt to delete every interval in the example.
+        The only things this guarantees to attempt to delete are every draw
+        call and every draw call plus its immediate successor (the first
+        non-empty draw call that starts strictly after it). However if this
+        seems to be working pretty well it will do its best to exploit that
+        and adapt to the fact there's currently a lot that it can delete.
 
         This is the main point at which we try to lower the size of the data.
         e.g. if we have two successive draw calls, this will attempt to delete
-        the first and replace it with the second. This also works for pairs of
-        draw calls (as we create a merged interval for adjacent draws at the
-        same level). So, for example, if we have something like:
+        the first and replace it with the second.
+
+        The fact that this will also try deleting the successor call is
+        important. For example, if we have something like:
 
         while many.more(data):
             data.draw(stuff)
@@ -1721,12 +1767,84 @@ class Shrinker(object):
         """
         self.debug('greedy interval deletes')
         i = 0
-        while i < len(self.intervals):
-            u, v = self.intervals[i]
-            if not self.incorporate_new_buffer(
-                self.shrink_target.buffer[:u] + self.shrink_target.buffer[v:]
-            ):
+        while i < len(self.shrink_target.examples):
+            if self.shrink_target.examples[i].length == 0:
                 i += 1
+                continue
+
+            # Note: We do want this fixed rather than changing during this
+            # iteration of the loop.
+            target = self.shrink_target
+
+            def try_delete_range(k):
+                """Can we delete k non-trivial non-overlapping examples
+                starting from i?"""
+                stack = []
+                j = i
+                while k > 0 and j < len(target.examples):
+                    ex = target.examples[j]
+                    if ex.length > 0 and (
+                        not stack or stack[-1][1] <= ex.start
+                    ):
+                        stack.append((ex.start, ex.end))
+                        k -= 1
+                    j += 1
+                assert stack
+                attempt = bytearray(target.buffer)
+                for u, v in reversed(stack):
+                    del attempt[u:v]
+                attempt = hbytes(attempt)
+                if sort_key(attempt) >= sort_key(self.shrink_target.buffer):
+                    return False
+                return self.incorporate_new_buffer(attempt)
+
+            # This is an adaptive pass loosely modelled after timsort. If
+            # little or nothing is deletable here then we don't try any more
+            # deletions than the naive greedy algorithm would, but if it looks
+            # like we have an opportunity to delete a lot then we try to do so.
+
+            # What we're trying to do is to find a large k such that we can
+            # delete k but not k + 1 draws starting from this point, and we
+            # want to do that in O(log(k)) rather than O(k) test executions.
+
+            # We try a quite careful sequence of small shrinks here before we
+            # move on to anything big. This is because if we try to be
+            # aggressive too early on we'll tend to find that we lose out when
+            # the example is "nearly minimal".
+            if try_delete_range(2):
+                if try_delete_range(3) and try_delete_range(4):
+
+                    # At this point it looks like we've got a pretty good
+                    # opportunity for a long run here. We do an exponential
+                    # probe upwards to try and find some k where we can't
+                    # delete many intervals. We do this rather than choosing
+                    # that upper bound to immediately be large because we
+                    # don't really expect k to be huge. If it turns out that
+                    # it is, the subsequent example is going to be so tiny that
+                    # it doesn't really matter if we waste a bit of extra time
+                    # here.
+                    hi = 5
+                    while try_delete_range(hi):
+                        assert hi <= len(target.examples)
+                        hi *= 2
+
+                    # We now know that we can delete the first lo intervals but
+                    # not the first hi. We preserve that property while doing
+                    # a binary search to find the point at which we stop being
+                    # able to delete intervals.
+                    lo = 4
+                    while lo + 1 < hi:
+                        mid = (lo + hi) // 2
+                        if try_delete_range(mid):
+                            lo = mid
+                        else:
+                            hi = mid
+            else:
+                try_delete_range(1)
+            # We unconditionally bump i because we have always tried deleting
+            # one more example than we succeeded at deleting, so we expect the
+            # next example to be undeletable.
+            i += 1
 
     def minimize_duplicated_blocks(self):
         """Find blocks that have been duplicated in multiple places and attempt
@@ -1864,9 +1982,9 @@ class Shrinker(object):
                 i += 1
 
     def interval_deletion_with_block_lowering(self):
-        """This pass tries to delete each interval (as per
-        greedy_interval_deletion) while replacing a block that precedes that
-        interval with its immediate two lexicographical predecessors.
+        """This pass tries to delete each interval while replacing a block that
+        precedes that interval with its immediate two lexicographical
+        predecessors.
 
         We only do this for blocks that are marked as shrinking - that
         is, when we tried lowering them it resulted in a smaller example.
